/*
 * Teensy 3.2 sketch to control ion grid and gas flow for the fusor.
 * PID control is interrupt driven, and controlled by flags from background loop, for ion supply control.
 * Gas solenoids are done in background loop, based on information from the PID control,
 * and all are controlled by flags from a host machine, including basic enable.
 * Data can be reported to the host via usb serial.
 * Using standard internal protocol, which goes like:
 * <unit number><command char><anything else up to 31 bytes total><lf>

 Existing commands: (preceeded by unit #, followed by <lf>, under 32 bytes, ffff = some float number)
 
 a - send "ack\n"
 
 cffff - set centerpoint for pid
 dffff - set differential coefficient for pid
 e - enable pid
 
 iffff - set integral coefficient for pid
 
 k - STOP and store tuning, clear variables
 l - led on
 o - led off
 
 pffff - set proportional coefficient
 
 r - report data
 s - save tuning in eeprom
 
 @@@ add solenoid times as tunable params, perhaps outlet till a condition instead of fixed time
 @@@ add code for valves
 */
 
#include <TimerOne.h> // for ISR
#include <ADC.h> // teensy adc
#include <EEPROM.h>

// defines
#define LED_PIN 13
#define NO_ONESHOTS 4
#define AD_DPin A9
#define AD_APin A2

// picked randomly as a place to start wearing eeprom
#define EEAddress 100
#define WINDUP 4000

// struct to save/restore with eeprom
struct Persist {
  float Pk;
  float Ik;
  float Dk;
  float Sp;
};

// Globals
Persist Tune; // where the PID vars live in ram

volatile byte oneshots[NO_ONESHOTS]; // put a number of ms in one, and it will count down to zero in ISR

ADC *adc = new ADC(); // the actual A/D object
volatile int AD_Desired; // setting for main HV current
volatile int AD_Actual; // actual current for HV
volatile float RawIonError; // convert to float right off for other calx
volatile float RawZ; // previous one for differential
volatile float pidD; // differential error
volatile float pidI; // integral error
volatile float pidF; // float output

volatile unsigned short DAOut; // value for D/A to put out


volatile boolean Hostenable; // various flags that can turn us off, this from host
volatile boolean HVenable; // this one from main HV supply
volatile boolean IonOK; // this one from ion supply
volatile boolean LEDState;

volatile unsigned int Punderflow;
volatile unsigned int Poverflow;

// above are the PID loop coefficients

// elapsedMicros usec = 0; // cool new volatile data type

int i;  // iterator for background stuff
int ii; // iterator for ISR stuff

// stuff for command state machine below

#define MY_ADDRESS '1'  // could be any byte, but we're using all-ascii for now, and '\n' for end of command
#define CMD_BUF_SIZE 32 // in theory, two of these would fit in the uno buffer at once.  Probably don't need this much IRL
#define MAX_TIME 20 // milliseconds to timeout if incomplete command (3ms would do for 115200 and fast commands)
byte Command[CMD_BUF_SIZE]; // command buffer for incoming commands
byte CommandIndex; // where am I in above buffer
byte CommandState; // state variable
unsigned long CommandTime; // time at start of command recieve
////// done with command-slave state machine stuff
//################################################################

void initRam() // used at startup and other times - be clean!
{
 Hostenable = HVenable = IonOK = false; // come up "off"
 CommandIndex = CommandState = 0;
 RawIonError = RawZ = pidI = 0; // pid stuff...
 DAOut = 0;
 for (i=0;i<NO_ONESHOTS;i++) oneshots[i] = 0; // clear all oneshots
 Punderflow = Poverflow = 0; // counts of pid loop stop hits
 EEPROM.get(EEAddress,Tune); // tuning saved from previous runs

}

//################################################################
void setup() {  // runs at reset
 pinMode(LED_PIN, OUTPUT); // set pin modes

 // set variables
 initRam();

 Serial.begin(115200); // use USB serial for host comm
 analogWriteResolution(12); // for built in D/A
 adc->setAveraging(8); // set number of averages
 adc->setResolution(12); // set bits of resolution
 adc->setSamplingSpeed(ADC_MED_SPEED); // change the sampling speed
 pinMode(AD_DPin, INPUT); // desired current input
 pinMode(AD_APin, INPUT); // actual current input

 Timer1.initialize(1000); // set up for 1ms interrupts
 Timer1.attachInterrupt(MsISR); // connect interrupt service routine
}

//########################## PID LOOP #############################
void MsISR() // hit once per ms, runs oneshots, PID control
{
 LEDState = true;
 digitalWrite(LED_PIN,LEDState);

  // count down any non-zero oneshots for use in loop()
 for (ii=0;ii<NO_ONESHOTS;ii++) if (oneshots[ii]) oneshots[ii]--;

 if (Hostenable /*&& HVenable && IonOK*/) // if everyone agrees...
 { // run pid
  AD_Desired = analogRead(AD_DPin);
  AD_Actual = analogRead(AD_APin);
  RawZ = RawIonError; // from last hit
  RawIonError = float(AD_Desired - AD_Actual); // raw error as float
  pidD = RawIonError - RawZ; // differential error
  pidI += RawIonError; // integrated error
  if (pidI > WINDUP) {pidI = WINDUP;} // prevent too much windup
  if (pidI < -WINDUP) {pidI = -WINDUP;}
  pidF = RawIonError * Tune.Pk + pidD * Tune.Dk + pidI * Tune.Ik + Tune.Sp; // raw pid output
  if (pidF > 4095)
  {
   pidF = 4095;
   Poverflow++;
  }
  if (pidF < 0)
  {
   pidF = 0;
   Punderflow++;
  }
  
  DAOut = int(pidF); 
 } else
 { // no pid, shut down clean
  
  DAOut = 0; // turn ion supply output off
 }
 
 analogWrite(A14,DAOut); // send whatever to D/A converter
 LEDState = false;
 digitalWrite(LED_PIN,LEDState);
}

//################################################################
void Report()
{ // report current status to host

 Serial.println("reporting as ordered, sir!");
 Serial.print ("DC: ");
 Serial.print (AD_Desired);
 Serial.print (" AC: ");
 Serial.print (AD_Actual);
 Serial.print (" Pk: ");
 Serial.print(Tune.Pk,4);
 Serial.print(" Ik: ");
 Serial.print(Tune.Ik,4);
 Serial.print(" Dk: ");
 Serial.print(Tune.Dk,4);
 Serial.print(" Sp: ");
 Serial.print(Tune.Sp);
 Serial.print(" of: ");
 Serial.print(Poverflow);
 Serial.print(" uf: ");
 Serial.print(Punderflow);
 Serial.println();
// Serial.print (" int size: ");
// Serial.println (sizeof(Punderflow)); 
}

//################################################################
float Str2Flt()
{ // parse string in Command[2] through end into float, return the float
 String numstring = ""; // to hold input number
 float value; // result

 i=2; // shared var, iterator/index here, 0 and 1 are unit number and command
 while (Command[i] != '\n') // \n is command end character
 {
//  Serial.print(char(Command[i])); // debug
  numstring += char(Command[i++]); // really does need cast to char
 }
// Serial.println(":done");
 value = numstring.toFloat();
 return value; // send it to caller
}

//################################################################
//////////////
void CommandMachine()
{
 switch (CommandState)
 {
  case 0: // looking for work to do
  if (Serial.available())
  {
   CommandTime = millis(); // for later possible timeout
   CommandState = 1; // get bytes till done or timeout
  }
  break;

  case 1: // getting command
  while (Serial.available())
  {
   Command[CommandIndex] = Serial.read(); // get a byte if available
//   Serial.print ("received:"); // debug
//   Serial.println(Command[CommandIndex]); // debug
   
   if (Command[CommandIndex] == '\n') // single quotes, else compiler thinks pointer to string
    { // got a full command, go parse it
     CommandState = 2;
     return; // get it on next roundy round of loop() 
    }
   CommandIndex++; // bump the position
   if ((CommandIndex >= CMD_BUF_SIZE) || ((CommandTime + MAX_TIME) < millis())) // enough parens?
   { // error of some kind, reset state machine
    CommandIndex = 0;
    CommandState = 0; // ignore what might have been garbage
//    Serial.println ("Command error"); // debug
   }
  } // end while
  break;
  
  case 2: // parse command
   if (Command[0] != MY_ADDRESS) // is this even for us?
   { // no, ditch it
    CommandIndex = 0;
    CommandState = 0;
    return; // forget it ever happened 
   }
   // It might have been for us, any legality tests might go after this comment.
   // By convention, address of unit is first byte, command is next, data might follow, should be enough.
   // For time consuming or big stuff, use a subroutine call instead of inlining it like this example.
   // Commands are all-ascii for convienience of host (and we slaves).
   switch (Command[1])
    {
     case 'a': // host requests an ack
        Serial.println("ack");  // should work
     break;
     case 'c': // set centerpoint
     Tune.Sp = Str2Flt();
     break;
     case 'd': // differential coefficient input
     Tune.Dk = Str2Flt();
     break;
     break; 
     case 'e': // enable pid etc
      Hostenable = true;
     break;
     case 'i': // input integral coeff, parse and store
     Tune.Ik = Str2Flt();
     break;
     case 'k': //host kills us
     Hostenable = false;
//@@@ clear out things here so start clean when (re)enabled
     EEPROM.put(EEAddress,Tune); // stash the pid variable struct
     break;     
     case 'l': // for test, led on
      digitalWrite(LED_PIN, HIGH);   // turn the LED on (HIGH is the voltage level)
     break; // end 'l' command
     case 'o': // led off command
      digitalWrite(LED_PIN, LOW);    // turn the LED off by making the voltage LOW
     break; // end off
    case 'p': // input of proportional coefficient, parse the float
    Tune.Pk = Str2Flt();
    break;
    case 'r': // report our "stuff"
     Report();
    break;
    case 's': // save our tuning in eeprom
     EEPROM.put(EEAddress,Tune); // stash the pid variable struct
    break;
    }    
  // we are done with this one   
     CommandIndex = 0;
     CommandState = 0;
//     Serial.println("end state 2"); // debug
    
   break; // end state 2
 } // end switch state machine
} // end command state machine routine


//################################################################



//################################################################
//                    MAIN
//################################################################

void loop() 
 { // main code here, to run repeatedly:

  CommandMachine();  // do things at request of host
  
//  digitalWrite(LED_PIN,LEDState); // show time through loop
//  LEDState  = !LEDState; // toggle it  
 }
